/* compilare con:
   make esempio_allocazione
   Eseguire, ad esempio, con
   ./esempio_allocazione < esempio_allocazione.txt
*/

#include <stdio.h> /* per printf */
#include <stdlib.h> /* per malloc e free */
#include <string.h> /* per memset */

#define DISPLAY_DATA(x) display_du(x,NUM_ELEMENTI); /* macro: tutti i DISPLAY_DATA(x)
						       vengono sostituiti da
						       display_du(x,NUM_ELEMENTI);
						       e poi NUM_ELEMENTI viene
						       sostituito da 10 */
#define NUM_ELEMENTI 10
#define MAX_LENGTH_NOME 20
#define BLOCK_SIZE 15

typedef struct {
  int peso;
  char sesso;
  char *cognome;
  char nome[MAX_LENGTH_NOME];
} datiutente;

void display_du(datiutente *, int);

void display_du(datiutente *du, int n){
  int i;
  for (i = 0; i < n; i++) {
    fflush(stdout);
    fflush(stderr);
    printf("N.%d\n\tpeso=>%d\n\tsesso=>%c\n\tnome=>%s\n\tcognome=>%s\n", i, du[i].peso, du[i].sesso,
    	   du[i].nome,
    	   du[i].cognome);
    fflush(stdout);
    fflush(stderr);
  }
}

int main (int argc, char *argv[]){
  datiutente *du[2];
  int i;
  /* alloca NUM_ELEMENTI*sizeof(datiutente) bytes, tutti inizializzati a zero
     ciascun blocco da sizeof(datiutente) bytes potra' poi essere acceduto
     come se fosse un elemento di un array il cui tipo base sia datiutente
  */
  du[0]=(datiutente *)calloc(NUM_ELEMENTI, sizeof(datiutente)); 
  if (du[0]==NULL) { // praticamente impossibile, si dovrebbe andare fuori dalla memoria virtuale...
    perror("impossibile allocare struttura dati datiutente");
    return -1;
  }
  for (i=0;i<NUM_ELEMENTI;i++) {
    char *tmp[2] = {NULL, NULL};
    int k;
    printf("N.%d - Inserisci peso, sesso, nome e cognome (separati da trattini)=>", i);
    /* legge da tastiera, a meno che non si ridiriga lo stdin */
    scanf("%d-%c-", &du[0][i].peso, &du[0][i].sesso);
    /* leggere nome e' facile: se ci sono troppi caratteri, non si ha altra
       scelta che tagliarli, dato che si tratta di un array a dimensione fissa. Ma
       cognome, essendo un puntatore generico, va gestito leggendo input di qualsiasi
       lunghezza. Siccome stdin non e' seekable (non si puo' tornare indietro e rileggere
       quanto gia' immesso), allora l'idea e' di provare a leggere a blocchi di dimensione
       BLOCK_SIZE, fintantoche' non si e' riusciti a leggere tutto (ovvero, finche' non
       si arriva al \n). Quindi, questo ciclo, all'iterazione i, alloca i*BLOCK_SIZE
       caratteri, rimpiazzando la memoria di (i - 1)*BLOCK_SIZE caratteri allocata
       all'iterazione precedente, grazie alla realloc
    */
    k = 1;
    do {
      tmp[0] = realloc(tmp[0], k*BLOCK_SIZE*sizeof(char));
      fgets(tmp[0] + (k - 1)*BLOCK_SIZE - (k > 1), BLOCK_SIZE + (k > 1), stdin);
      k++;
    } while(tmp[0][strlen(tmp[0]) - 1] != '\n');
    /* un trattino (-) separa nome da cognome; del nome si prendono solo
       (MAX_LENGTH_NOME - 2) caratteri (quello ulteriore sara' la terminazione
       stringa) */
    tmp[1] = strtok(tmp[0], "-"); /* strtok separa la stringa data come primo argomento
				     sulla base del secondo argomento: un po' come la
				     funzione split di Python e Java. Solo che non torna
				     un array, ma un puntatore alla prima stringa separata
				     (token); vedere esempio_strtok.c per un esempio
				     minimale */
    /* provare a compilare con:

       make CFLAGS="-ggdb" esempio_allocazione

       e poi debuggare passo passo con

       gdb esempio_allocazione

       Ad esempio, dalla shell di gdb, si possono dare i seguenti comandi:

       b 79
       run < esempio_allocazione.txt

       il primo setta un breakpoint all'istruzione subito sopra (quella con la strtok),
       il secondo effettua l'esecuzione del programma fino al breakpoint (sempre con lo stdin
       rediretto)
       A questo punto, visualizzare il valore di tmp[0] con:

       p tmp[0]

       come si vede, contiene l'intera riga letta; ora dare i comandi:

       next
       p tmp[0]

       ora non e' piu' l'intera riga, ma solo la prima parte (fino al trattino escluso).
       Nota bene: in realta' tmp[0] punta ancora all'intera zona di memoria; semplicemente,
       strtok ha sostituito il trattino con \0. Per vederlo, si puo' usare la seguente
       sintassi:

       p tmp[0][0]@74

       in questo modo, si sta dicendo a gdb di trattare tmp[0] come un array (di caratteri),
       e di stamparne i primi 74
    */
    strncpy(du[0][i].nome, tmp[1], MAX_LENGTH_NOME - 1);
    if (strlen(tmp[1]) > MAX_LENGTH_NOME - 1)
      du[0][i].nome[MAX_LENGTH_NOME - 1] = '\0';
    tmp[1] = strtok(NULL, "-"); /* qui si ottiene la seconda: tutte le chiamate a strtok
				   successive alla prima hanno come primo argomento NULL */
    /* L'effetto della strtok di prima, vista a livello di memoria, e' quello di assegnare
       al puntatore tmp[1] l'indirizzo di memoria immediatamente seguente il trattino
       (che ora e' sostituito dal \0, vedere sopra) */
    /* du[0][i].cognome punta all'interno dell'ultimo blocco allocato da realloc */
    du[0][i].cognome = tmp[1];
    tmp[1][strlen(tmp[1]) - 1] = '\0';
  }
  printf("I dati inseriti dall'utente sono:\n");
  DISPLAY_DATA(du[0]); //punto e virgola inutile (c'e' gia' nella macro), messo solo perche' altrimenti l'editor fa storie con l'indentazione
  du[1]=(datiutente *)malloc(NUM_ELEMENTI * sizeof(datiutente));
  memcpy(du[1], du[0], NUM_ELEMENTI * sizeof(datiutente));
  printf("du[0] e du[1] %ssono uguali\n", memcmp(du[0], du[1], NUM_ELEMENTI * sizeof(datiutente)) == 0? "" : "non ");
  printf("Contenuto du[0] dopo memset\n");
  memset(du[0], 0, NUM_ELEMENTI*sizeof(datiutente));
  DISPLAY_DATA(du[0]);
  printf("du[0] e du[1] %ssono uguali\n", memcmp(du[0], du[1], NUM_ELEMENTI * sizeof(datiutente)) == 0? "" : "non ");
  free(du[0]); //praticamente inutile, il programma e' finito, e non sta liberando tutto!
               //mancano infatti tutti i .cognome; valgrind se ne accorge...
  free(du[1]); //praticamente inutile, idem
  return 0;
}

/*

  esercizio 1: modificare in modo da liberare tutta la memoria allocata (controllare con
  valgrind --leak-check=full)

  esercizio 2: modificare assumendo che l'input sia da file, eliminando la realloc
  
 */
